
/*****************************************************************************************************************************
--
--    File Name    : sharpen.v 

--    Description  : This module implements Unsharp Mask .

-- Targeted device : Microsemi-SoC                     
-- Author          : India Solutions Team

-- SVN Revision Information:
-- SVN $Revision: TBD
-- SVN $Date: TBD
--
--
--
-- COPYRIGHT 2015 BY MICROSEMI 
-- THE INFORMATION CONTAINED IN THIS DOCUMENT IS SUBJECT TO LICENSING RESTRICTIONS 
-- FROM MICROSEMI CORP.  IF YOU ARE NOT IN POSSESSION OF WRITTEN AUTHORIZATION FROM 
-- MICROSEMI FOR USE OF THIS FILE, THEN THE FILE SHOULD BE IMMEDIATELY DESTROYED AND 
-- NO BACK-UP OF THE FILE SHOULD BE MADE. 
-- 

--******************************************************************************************************************************/

module sharpen #(parameter g_DATA_BITWIDTH = 8)
               (input wire SYS_CLK_I,
                input wire nRESET_I,
                input wire[g_DATA_BITWIDTH - 1 : 0] p0_i,
                input wire[g_DATA_BITWIDTH - 1 : 0] p1_i,
                input wire[g_DATA_BITWIDTH - 1 : 0] p2_i,
                input wire[g_DATA_BITWIDTH - 1 : 0] p3_i,
                input wire[g_DATA_BITWIDTH - 1 : 0] p4_i,
                input wire[g_DATA_BITWIDTH - 1 : 0] p5_i,
                input wire[g_DATA_BITWIDTH - 1 : 0] p6_i,
                input wire[g_DATA_BITWIDTH - 1 : 0] p7_i,
                input wire[g_DATA_BITWIDTH - 1 : 0] p8_i,
                output wire[g_DATA_BITWIDTH - 1 : 0] DATA_OUT_o);
  reg[g_DATA_BITWIDTH + 1 : 0] Oil;
  reg[g_DATA_BITWIDTH + 1 : 0] Iil;
  reg[g_DATA_BITWIDTH - 1 : 0] lil;
  reg signed[g_DATA_BITWIDTH + 3 : 0] oil;
  reg[g_DATA_BITWIDTH - 1 : 0] iil;
  reg signed[g_DATA_BITWIDTH + 4 : 0] OO0;
  always
    @(posedge (SYS_CLK_I),
      negedge (nRESET_I))
    begin
      if (!nRESET_I)
        begin
          Oil <= { g_DATA_BITWIDTH + 2 { 1'b0 } };
        end
      else
        begin
          Oil <= p0_i + p1_i + p2_i + p3_i;
        end
    end
  always
    @(posedge (SYS_CLK_I),
      negedge (nRESET_I))
    begin
      if (!nRESET_I)
        begin
          Iil <= { g_DATA_BITWIDTH + 2 { 1'b0 } };
        end
      else
        begin
          Iil <= p5_i + p6_i + p7_i + p8_i;
        end
    end
  always
    @(posedge (SYS_CLK_I),
      negedge (nRESET_I))
    begin
      if (!nRESET_I)
        begin
          lil <= { g_DATA_BITWIDTH { 1'b0 } };
        end
      else
        begin
          lil <= p4_i;
        end
    end
  always
    @(posedge (SYS_CLK_I),
      negedge (nRESET_I))
    begin
      if (!nRESET_I)
        begin
          oil <= { g_DATA_BITWIDTH + 4 { 1'b0 } };
        end
      else
        begin
          oil <= 8 * lil - (Oil + Iil);
        end
    end
  always
    @(*)
    begin
      if (oil[g_DATA_BITWIDTH + 3])
        begin
          OO0 <= ~oil + 1;
        end
      else
        begin
          OO0 <= { oil[g_DATA_BITWIDTH + 3], oil };
        end
    end
  always
    @(posedge (SYS_CLK_I),
      negedge (nRESET_I))
    begin
      if (!nRESET_I)
        begin
          iil <= { g_DATA_BITWIDTH { 1'b0 } };
        end
      else
        begin
          iil <= (OO0[g_DATA_BITWIDTH - 1 : 0]);
        end
    end
  assign DATA_OUT_o = iil[g_DATA_BITWIDTH - 1 : 0];
endmodule
