/*****************************************************************************************************************************
--
--    File Name    : display_controller_test.v

--    Description  : Top-level Verilog TB for display controller
--                   

-- Targeted device : Microsemi-SoC
-- Author          : India Solutions Team

-- SVN Revision Information:
-- SVN $Revision: TBD
-- SVN $Date: TBD
--
--
--
-- COPYRIGHT 2016 BY MICROSEMI 
-- THE INFORMATION CONTAINED IN THIS DOCUMENT IS SUBJECT TO LICENSING RESTRICTIONS 
-- FROM MICROSEMI CORP.  IF YOU ARE NOT IN POSSESSION OF WRITTEN AUTHORIZATION FROM 
-- MICROSEMI FOR USE OF THIS FILE, THEN THE FILE SHOULD BE IMMEDIATELY DESTROYED AND 
-- NO BACK-UP OF THE FILE SHOULD BE MADE. 
-- 

--******************************************************************************************************************************/
`timescale 1ps/100fs
module display_controller_test();

`define VIDEO_CLOCK_TP 		6734	 		//Default : 71MHz
`define VIDEO_CLOCK_HALF_TP 	`VIDEO_CLOCK_TP/2
`define FRAMES_PER_SEC 		60  			//60fps


  /************************************************************************
          Parameters 
  *************************************************************************/

 parameter SYS_CLOCK_PERIOD 	    		= 6252	; 
 parameter g_DISPLAY_RESOLUTION 		= 320	;
 parameter g_VERT_DISPLAY_RESOLUTION 		= 200	;
 parameter IMAGE_FILE_NAME = "RGB_in.txt";  

 //Parameters from the DUT
 parameter g_DDR_AXI_AWIDTH 			= 32	;
 parameter g_DEPTH_OF_VIDEO_PIXEL_FROM_DDR 	= 1	;
 parameter g_HALF_EMPTY_THRESHOLD 		= 960 	;
 parameter g_HORZ_SYNC_PULSE_POLARITY 		= 1	;
 parameter g_INITIAL_LINES_TO_BUFFER 		= 6	;
 parameter g_INPUT_VIDEO_DATA_BIT_WIDTH 	= 24	;
 parameter g_INPUT_X_W_RES_WIDTH 		= 9	;
 parameter g_INPUT_Y_H_RES_WIDTH 		= 8	;
 parameter g_SUBSEQUENT_LINES_TO_BUFFER 	= 1	;
 parameter g_VERT_SYNC_PULSE_POLARITY 		= 1	;
 parameter g_VIDEO_FIFO_AWIDTH 			= 12	;

 parameter SIZE = g_VERT_DISPLAY_RESOLUTION * g_DISPLAY_RESOLUTION *3 ; 
 parameter SYS_CLOCK_HALF_TP 	    		= SYS_CLOCK_PERIOD/2	;

   /************************************************************************
          Inputs and Outputs to the DUT 
  *************************************************************************/
  reg						     RESET_N_I			;		
  reg						     SYS_CLOCK_I		;		
  reg						     VIDEO_CLOCK_I		;		
  reg						     display_control_enable_i	;		
  reg	    [(g_INPUT_VIDEO_DATA_BIT_WIDTH - 1) : 0] video_datain_i		;		
  reg						     data_valid_i		;		
  reg	    [(g_DDR_AXI_AWIDTH - 1) : 0]	     frame_buff_addr_i		;		
  reg						     frame_ddr_addr_valid_i	;		
  reg						     read_ack_i			;		
  reg						     read_done_i		;		
  reg	    [(g_INPUT_X_W_RES_WIDTH - 1) : 0]	     tot_horz_pixels_i		;		
  reg	    [(g_INPUT_Y_H_RES_WIDTH - 1) : 0]	     tot_vert_lines_i		;		
  reg	    [31 : 0]				     video_resolution_i		;		
  reg	    [31 : 0]				     back_porch_width_i		;		
  reg	    [31 : 0]				     front_porch_width_i	;		
  reg	    [(g_INPUT_X_W_RES_WIDTH - 1) : 0]	     hsync_width_i		;		
  reg	    [(g_INPUT_Y_H_RES_WIDTH - 1) : 0]	     vsync_width_i		;		
  reg	    [31 : 0]	  			     ddr_rd_video_resolution_i	;		
  wire						     disp_cont_busy_o		;		
  wire						     disp_cont_done_o		;		
  wire	    [(g_DDR_AXI_AWIDTH - 1) : 0]	     frame_buff_und_disp_o	;		
  wire	    [15 : 0]				     bytes_to_read_o		;		
  wire	    [(g_DDR_AXI_AWIDTH - 1) : 0]	     start_read_addr_o		;		
  wire						     read_req_o			;		
  wire	    [23 : 0]				     rgb_out_o			;		
  wire						     data_enable_o		;		
  wire						     horz_sync_o		;		
  wire						     vert_sync_o		;		
  wire						     hactive_o			;		
  wire						     vactive_o			;		
  wire						     dg_fifo_empty		;
  wire						     dg_fifo_full		;
  wire						     dg_fifo_hempty		;		

  reg  	   [7:0]				     image_read[((SIZE-1)+4):0]	; 
  reg  	   [7:0]				     image_input[SIZE-1:0]	; 
  reg  	   [7:0]				     img_out_array[SIZE-1:0]	; 
  reg 						     img_proc_started 	= 1'b0	;
  reg      [31:0]                                    disp_resolution	               ;
  reg      [15:0]                                    horz_resolution	               ;
  reg      [15:0]                                    vert_resolution	               ;

  integer frame_handle ;
  integer pixels_read = 0;
  integer in_byte_cnt = 0;
  integer img_out_pixel_cnt = 0;
  integer no_of_pixels_per_sec =1;
  integer video_clock_time_period_ns	=`VIDEO_CLOCK_TP ;
  integer video_clock_half_time_period	=`VIDEO_CLOCK_HALF_TP; 
  integer  i=0;


  /************************************************************************
          Read the input image 
  *************************************************************************/
  initial
    begin
      $readmemh(IMAGE_FILE_NAME, image_read);
      for (i = 0; i < SIZE; i = i + 1)
        begin
          image_input[i] =  image_read[i  + 4];
        end
      $display("");
      $display("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
      $display(">>>>>>>>                     IMAGE LOADED                                           <<<<<<<");
      $display("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
    end

 /************************************************************************
          Generates System Clock signal
  *************************************************************************/	
 initial
   begin
     SYS_CLOCK_I = 1'b0;
     while(1)begin
	#(SYS_CLOCK_HALF_TP) SYS_CLOCK_I = ~SYS_CLOCK_I;
     end
   end

 /************************************************************************
          Generates Video Clock signal
  *************************************************************************/
  initial
    begin
	VIDEO_CLOCK_I = 1'b0;
	while(1)begin
		#(video_clock_half_time_period) VIDEO_CLOCK_I = ~VIDEO_CLOCK_I;
	end
    end


 /************************************************************************
          RESET : Generates a reset pulse 
  *************************************************************************/
  initial
    begin
	RESET_N_I = 1'b1;
	#10000 RESET_N_I = 1'b0;
	#30000 RESET_N_I = 1'b1;
    end

 /************************************************************************
  Activate reset conditions and drive applicable timing parameter constants
  *************************************************************************/

  initial
    begin
	forever begin 
		@ (negedge RESET_N_I);
		reset();
	        $display("");
	        $display("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
	        $display(">>>>>>>>                    RESET APPLIED		                            <<<<<<<");
	        $display("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
      end
    end

 /************************************************************************
  Generating constants according to input resolution
  *************************************************************************/

  initial
    begin
	forever begin 
		@ (posedge RESET_N_I);
		generate_timing_constants();
		print_constants();
	        $display("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
	        $display(">>>>>>>>                    CONSTANTS REFRESHED     				    <<<<<<<");
	        $display("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
		calc_vid_clk_period();
		display_control_enable_i	= 1'b1;
	end
    end

  /************************************************************************
  Task to calculate video clock period according to latest timing parameter constants
  *************************************************************************/

  task calc_vid_clk_period();
    begin
	no_of_pixels_per_sec = tot_horz_pixels_i*tot_vert_lines_i*`FRAMES_PER_SEC ;
	video_clock_time_period_ns = 1000000000/no_of_pixels_per_sec ; 			       
	video_clock_half_time_period =  1000*video_clock_time_period_ns/2;
	$display("video_clock_time_period in nanosec is %d ",video_clock_time_period_ns);
    end
  endtask

  /************************************************************************
          Task to reset signals
  *************************************************************************/

  task reset();
    begin
	display_control_enable_i	= 1'b0;
	frame_buff_addr_i		= 32'hADC00000;
	frame_ddr_addr_valid_i		= 1'b0;
	video_datain_i			= 32'b0;
	data_valid_i			= 1'b0;
	read_ack_i			= 1'b0;
	read_done_i			= 1'b0;
    end
  endtask

 /************************************************************************
          Updating ddr_rd_video_resolution_i input
  *************************************************************************/	
 
 always @ (posedge SYS_CLOCK_I,  negedge RESET_N_I) 
   begin
	   if (RESET_N_I == 0) begin
		   ddr_rd_video_resolution_i 	= 32'h01140140;
	   end
	   else begin
		   ddr_rd_video_resolution_i 	= video_resolution_i;
	   end
   end

  /************************************************************************
  Response to read request from dut
  *************************************************************************/

  initial
    begin
	while (1) begin
		@ (posedge SYS_CLOCK_I) ;
		if (read_req_o ==1 ) begin
			if (img_proc_started==1'b0)begin
				img_proc_started = 1'b1;
				$display("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
				$display(">>>>>>>>                            IMAGE PROCESSING STARTED                        <<<<<<<");
				$display("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
			end
			read_ack_i =1'b1;
			@ (posedge SYS_CLOCK_I) ;
			read_ack_i =1'b0;
			write_image_data_to_dc();
			read_done_i = 1'b1 ;
			@ (posedge SYS_CLOCK_I) ;
			read_done_i = 1'b0 ;
        	end
	end
    end

  /************************************************************************
  Task to Send Valid image data input depending on read request
  *************************************************************************/

  task write_image_data_to_dc(); 
    begin
	for (in_byte_cnt=0;in_byte_cnt<bytes_to_read_o;in_byte_cnt= in_byte_cnt +1)	begin
		video_datain_i  = { image_input[pixels_read*3], image_input[pixels_read*3 +1], image_input[pixels_read*3 +2]};
		pixels_read =  pixels_read +1;
		data_valid_i    = 1'b1;
		@ (posedge SYS_CLOCK_I);
	end	
	data_valid_i    = 1'b0;
    end
  endtask


  /************************************************************************
  Process to write the output RGB into file
  *************************************************************************/
  initial
    begin
	frame_handle = $fopen("frame_out.txt","wb");  
	$fwrite(frame_handle,"\n\n\n");
	while(1)begin
		if (data_enable_o == 1'b1)begin
			$fwrite(frame_handle,"%x\n", rgb_out_o[23:16]);
			$fwrite(frame_handle,"%x\n", rgb_out_o[15:8]);
			$fwrite(frame_handle,"%x\n", rgb_out_o[7:0]);
			img_out_pixel_cnt = img_out_pixel_cnt +1;
		end
		@ (posedge VIDEO_CLOCK_I) ; 
	end
  end

 /************************************************************************
 	Closing Output image File and completing simulation
  *************************************************************************/

  initial 
    begin
	@ (posedge vert_sync_o) ; 
	$fclose(frame_handle);
        $display("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
        $display(">>>>>>>>                             SIMULATION COMPLETED                           <<<<<<<");
        $display("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
	#1000000;
	$finish;
   end


/*****************************************************************************
	Task  to select timing parameter constants based on Resolution
*****************************************************************************/

  task generate_timing_constants();
    begin
	horz_resolution = g_DISPLAY_RESOLUTION ; 
	vert_resolution = g_VERT_DISPLAY_RESOLUTION ; 
	disp_resolution = {horz_resolution,vert_resolution} ; 

	case(disp_resolution)
		{16'd1280,16'd800} : begin
				tot_horz_pixels_i   = 1440;					
				tot_vert_lines_i    = 823;					
				video_resolution_i  = {{16'd800}, {16'd1280}};			
				back_porch_width_i  = 32'h000E0050;
				front_porch_width_i = 32'h00030030;				
				hsync_width_i       = 12'h020;					
				vsync_width_i       = 12'h006;					
		end
		{16'd1280,16'd720} : begin
				tot_horz_pixels_i   = 1440;                                     
				tot_vert_lines_i    = 741;                                      
				video_resolution_i  = {{16'd720},{16'd1280}};	                
				back_porch_width_i  = 32'h000D0050;
				front_porch_width_i = 32'h00030030;                             
				hsync_width_i       = 12'h020;                                  
				vsync_width_i       = 12'h005;                                  
		end
		{16'd1920,16'd1080} : begin
				tot_horz_pixels_i   = 2200;                                     
				tot_vert_lines_i    = 1125;                                     
				video_resolution_i  = {{16'd1080},{16'd1920}};	                
				back_porch_width_i  = 32'h00240094;
				front_porch_width_i = 32'h00040058;                             
				hsync_width_i       = 12'h02C;                                  
				vsync_width_i       = 12'h005;                                  
		end
		{16'd2048,16'd1080} : begin
				tot_horz_pixels_i   = 2208;                                     
				tot_vert_lines_i    = 1111;                                     
				video_resolution_i  = {{16'd1080},{16'd2048}};	                
				back_porch_width_i  = 32'h00120050;
				front_porch_width_i = 32'h00030030;                             
				hsync_width_i       = 12'h020;                                  
				vsync_width_i       = 12'h00A;                                  
		end

		{16'd2560,16'd1600}  : begin
				tot_horz_pixels_i   = 2720;                                     
				tot_vert_lines_i    = 1646;                                     
				video_resolution_i  = {{16'd1600},{16'd2560}};	                
				back_porch_width_i  = 32'h00250050;
				front_porch_width_i = 32'h00030030;                             
				hsync_width_i       = 12'h020;                                  
				vsync_width_i       = 12'h006;                                  
		end
		{16'd3840,16'd2160} : begin
				tot_horz_pixels_i   = 4000;                                     
				tot_vert_lines_i    = 2222;                                     
				video_resolution_i  = {{16'd2160},{16'd3840}};	                
				back_porch_width_i  = 32'h00360050;
				front_porch_width_i = 32'h00030030;                             
				hsync_width_i       = 12'h020;                                  
				vsync_width_i       = 12'h005;                                  
		end
		{16'd4096,16'd2160} : begin
				tot_horz_pixels_i   = 4256;                                     
				tot_vert_lines_i    = 2222;                                     
				video_resolution_i  = {{16'd2160},{16'd4096}};	                
				back_porch_width_i  = 32'h00310050;
				front_porch_width_i = 32'h00030030;                             
				hsync_width_i       = 12'h020;                                  
				vsync_width_i       = 12'h00A;                                  
		end
		{16'd800,16'd600} : begin
				tot_horz_pixels_i   = 960;                                      
				tot_vert_lines_i    = 618;                                      
				video_resolution_i  = {{16'd600},{16'd800}};	                
				back_porch_width_i  = 32'h000B0050;
				front_porch_width_i = 32'h00030030;                             
				hsync_width_i       = 12'h020;                                  
				vsync_width_i       = 12'h004;                                  
		end
		{16'd600,16'd380} : begin
				tot_horz_pixels_i   = 760;                                      
				tot_vert_lines_i    = 399;                                      
				video_resolution_i  = {{16'd380},{16'd600}};	                
				back_porch_width_i  = 32'h00060050;
				front_porch_width_i = 32'h00030030;                             
				hsync_width_i       = 12'h020;                                  
				vsync_width_i       = 12'h00A;                                  
		end
		{16'd320,16'd280} : begin
				tot_horz_pixels_i   = 480;                                      
				tot_vert_lines_i    = 299;                                      
				video_resolution_i  = {{16'd280},{16'd320}};	                
				back_porch_width_i  = 32'h00060050;
				front_porch_width_i = 32'h00030030;                             
				hsync_width_i       = 12'h020;                                  
				vsync_width_i       = 12'h00A;                                  
		end
		{16'd320,16'd200} : begin
				tot_horz_pixels_i   = 480;                                      
				tot_vert_lines_i    = 215;                                      
				video_resolution_i  = {{16'd200},{16'd320}};	                
				back_porch_width_i  = 32'h00060050;
				front_porch_width_i = 32'h00030030;                             
				hsync_width_i       = 12'h020;                                  
				vsync_width_i       = 12'h006;                                  
		end
		default		 : begin
				tot_horz_pixels_i   = 480;                                      
				tot_vert_lines_i    = 299;                                      
				video_resolution_i  = {{16'd280},{16'd320}};	                
				back_porch_width_i  = 32'h00060050;
				front_porch_width_i = 32'h00030030;                             
				hsync_width_i       = 12'h020;                                  
				vsync_width_i       = 12'h00A;
				$display("Assigning default values");			
		end
      endcase
	$display("TIMING PARAMETER CONSTANTS UPDATED ACCORDING TO CHOSEN DISPLAY RESOLUTION : x%h ",disp_resolution);
	$display(" DISPLAY HORIZONTAL RESOLUTION : %d,  DISPLAY VERTICAL RESOLUTION : %d ",horz_resolution, vert_resolution);
    end
  endtask

/*****************************************************************************
	Task  to print timing parameter constants selected
*****************************************************************************/
  task print_constants();
    begin
	$display(" \n	CONSTANTS CHOSEN AS TIMING PARAMETERS ARE :\n"			 )     ;
	$display(" 	tot_horz_pixels_i	:	 %d	    ",tot_horz_pixels_i	 )     ;						
	$display(" 	tot_vert_lines_i	:	 %d	    ",tot_vert_lines_i	 )     ;						
	$display(" 	video_resolution_i	:	x%h	    ",video_resolution_i )     ;				
	$display(" 	back_porch_width_i	:	x%h	    ",back_porch_width_i )     ;
	$display(" 	front_porch_width_i	:	x%h	    ",front_porch_width_i)     ;					
	$display(" 	hsync_width_i		:	x%h	    ",hsync_width_i	 )     ;						
	$display(" 	vsync_width_i		:	x%h	    ",vsync_width_i	 )     ;						
    end
  endtask


  /************************************************************************
  Component instantiation
  *************************************************************************/

 display_controller # (	   .g_DDR_AXI_AWIDTH			(g_DDR_AXI_AWIDTH)			,
                           .g_INPUT_X_W_RES_WIDTH		(g_INPUT_X_W_RES_WIDTH)			,
                           .g_INPUT_Y_H_RES_WIDTH		(g_INPUT_Y_H_RES_WIDTH)			,
                           .g_VIDEO_FIFO_AWIDTH			(g_VIDEO_FIFO_AWIDTH)			,
                           .g_INPUT_VIDEO_DATA_BIT_WIDTH	(g_INPUT_VIDEO_DATA_BIT_WIDTH)		,
                           .g_DEPTH_OF_VIDEO_PIXEL_FROM_DDR	(g_DEPTH_OF_VIDEO_PIXEL_FROM_DDR)	,
                           .g_HORZ_SYNC_PULSE_POLARITY		(g_HORZ_SYNC_PULSE_POLARITY)		,
                           .g_VERT_SYNC_PULSE_POLARITY		(g_VERT_SYNC_PULSE_POLARITY)		,
						   .g_INITIAL_LINES_TO_BUFFER		(g_INITIAL_LINES_TO_BUFFER)		,
						   .g_SUBSEQUENT_LINES_TO_BUFFER	(g_SUBSEQUENT_LINES_TO_BUFFER)		,
                           .g_HALF_EMPTY_THRESHOLD		(g_HALF_EMPTY_THRESHOLD)

					) 
displ_ctrl_inst		(
                           
						   .RESET_N_I			(RESET_N_I)                                   ,
                           .SYS_CLOCK_I			(SYS_CLOCK_I)                                 ,
                           .VIDEO_CLOCK_I		(VIDEO_CLOCK_I)                               ,
                           .display_control_enable_i	(display_control_enable_i)                    ,
                           .video_datain_i		(video_datain_i)                              ,
                           .data_valid_i		(data_valid_i)                                ,
                           .frame_buff_addr_i		(frame_buff_addr_i)                           ,
                           .frame_ddr_addr_valid_i	(frame_ddr_addr_valid_i)                      ,
                           .read_ack_i			(read_ack_i)                                  ,
                           .read_done_i			(read_done_i)                                 ,
                           .tot_horz_pixels_i		(tot_horz_pixels_i)                           ,
                           .tot_vert_lines_i		(tot_vert_lines_i)                            ,
                           .video_resolution_i		(video_resolution_i)                          ,
                           .back_porch_width_i		(back_porch_width_i)                          ,
                           .front_porch_width_i		(front_porch_width_i)                         ,
                           .hsync_width_i		(hsync_width_i)                               ,
                           .vsync_width_i		(vsync_width_i)                               ,
                           .ddr_rd_video_resolution_i	(ddr_rd_video_resolution_i)                   ,
                           
						   .disp_cont_busy_o		(disp_cont_busy_o)                            ,
                           .disp_cont_done_o		(disp_cont_done_o)                            ,
                           .frame_buff_und_disp_o	(frame_buff_und_disp_o)                       ,
                           .bytes_to_read_o		(bytes_to_read_o)                             ,
                           .start_read_addr_o		(start_read_addr_o)                           ,
                           .read_req_o			(read_req_o)                                  ,
                           .rgb_out_o			(rgb_out_o)                                   ,
                           .data_enable_o		(data_enable_o)                               ,
                           .horz_sync_o			(horz_sync_o)                                 ,
                           .vert_sync_o			(vert_sync_o)                                 ,
                           .hactive_o			(hactive_o)                                   ,
                           .vactive_o			(vactive_o)                                   ,
                           .dg_fifo_empty		(dg_fifo_empty)                               ,
						   .dg_fifo_full		(dg_fifo_full)                                ,
                           .dg_fifo_hemtpy		(dg_fifo_hempty)                              
                        );



endmodule //display_controller_test
