/*****************************************************************************************************************************
--
--    File Name    : embsync_detect.v 

--    Description  : This module implements detection of MIPI Sync code.


-- Targeted device : Microsemi-SoC                     
-- Author          : India Solutions Team

-- SVN Revision Information:
-- SVN $Revision: TBD
-- SVN $Date: TBD
--
--
--
-- COPYRIGHT 2018 BY MICROSEMI 
-- THE INFORMATION CONTAINED IN THIS DOCUMENT IS SUBJECT TO LICENSING RESTRICTIONS 
-- FROM MICROSEMI CORP.  IF YOU ARE NOT IN POSSESSION OF WRITTEN AUTHORIZATION FROM 
-- MICROSEMI FOR USE OF THIS FILE, THEN THE FILE SHOULD BE IMMEDIATELY DESTROYED AND 
-- NO BACK-UP OF THE FILE SHOULD BE MADE. 
-- 

--******************************************************************************************************************************/
module embsync_detect #(parameter g_LANE_WIDTH = 4)
                      (input l1l,
                       input CLOCK_I,
                       input HS,
                       input[7 : 0] L0_DATA_I,
                       input[7 : 0] L1_DATA_I,
                       input[7 : 0] L2_DATA_I,
                       input[7 : 0] L3_DATA_I,
                       output[7 : 0] L0_DATA_O,
                       output[7 : 0] L1_DATA_O,
                       output[7 : 0] L2_DATA_O,
                       output[7 : 0] L3_DATA_O);
  reg[7 : 0] o1l,
             i1l,
             Ool,
             Iol;
  reg[6 : 0] lol,
             ool,
             iol,
             Oil;
  reg[2 : 0] Iil,
             lil,
             oil,
             iil;
  reg[7 : 0] OO0,
             IO0,
             lO0,
             oO0;
  wire[7 : 0] iO0,
              OI0,
              II0,
              lI0;
  reg oI0,
      iI0,
      Ol0,
      Il0;
  reg ll0,
      ol0,
      il0,
      O00;
  reg[7 : 0] I00,
             l00,
             o00,
             i00,
             O10,
             I10,
             l10,
             o10;
  reg[7 : 0] i10,
             Oo0,
             Io0,
             lo0,
             oo0,
             io0,
             Oi0,
             Ii0;
  reg[7 : 0] li0,
             oi0,
             ii0,
             OO1,
             IO1,
             lO1,
             oO1,
             iO1;
  reg[7 : 0] OI1,
             II1,
             lI1,
             oI1,
             iI1,
             Ol1,
             Il1,
             ll1;
  reg[7 : 0] ol1,
             il1,
             O01,
             I01,
             l01,
             o01,
             i01,
             O11;
  reg[7 : 0] I11,
             l11,
             o11,
             i11,
             Oo1,
             Io1,
             lo1,
             oo1;
  reg[7 : 0] io1,
             Oi1,
             Ii1,
             li1,
             oi1,
             ii1,
             OOo,
             IOo;
  reg[7 : 0] lOo,
             oOo,
             iOo,
             OIo,
             IIo,
             lIo,
             oIo,
             iIo;
  reg Olo,
      Ilo,
      llo,
      olo,
      ilo;
  reg O0o;
  reg[10 : 0] I0o;
  parameter l0o = 8'b00011101;
  always
    @(posedge CLOCK_I or
      negedge l1l)
    if (!l1l)
      begin
        Olo <= 0;
        Ilo <= 0;
        llo <= 0;
        olo <= 0;
        ilo <= 0;
      end
    else
      begin
        Olo <= HS;
        Ilo <= Olo;
        llo <= Ilo;
        olo <= llo;
        ilo <= olo;
      end
  generate
    if (g_LANE_WIDTH > 0)
      begin
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              begin
                o1l <= 0;
                lol <= 0;
              end
            else
              begin
                o1l <= { L0_DATA_I[0], L0_DATA_I[1], L0_DATA_I[2], L0_DATA_I[3], L0_DATA_I[4], L0_DATA_I[5], L0_DATA_I[6], L0_DATA_I[7] };
                lol <= o1l[6 : 0];
              end
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              begin
                I00 <= 0;
                i10 <= 0;
                li0 <= 0;
                OI1 <= 0;
                ol1 <= 0;
                I11 <= 0;
                io1 <= 0;
                lOo <= 0;
                l00 <= 0;
                Oo0 <= 0;
                oi0 <= 0;
                II1 <= 0;
                il1 <= 0;
                l11 <= 0;
                Oi1 <= 0;
                oOo <= 0;
              end
            else
              begin
                I00 <= o1l;
                i10 <= { lol[0], o1l[7 : 1] };
                li0 <= { lol[1 : 0], o1l[7 : 2] };
                OI1 <= { lol[2 : 0], o1l[7 : 3] };
                ol1 <= { lol[3 : 0], o1l[7 : 4] };
                I11 <= { lol[4 : 0], o1l[7 : 5] };
                io1 <= { lol[5 : 0], o1l[7 : 6] };
                lOo <= { lol[6 : 0], o1l[7] };
                l00 <= I00;
                Oo0 <= i10;
                oi0 <= li0;
                II1 <= OI1;
                il1 <= ol1;
                l11 <= I11;
                Oi1 <= io1;
                oOo <= lOo;
              end
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              begin
                Iil <= 0;
                oI0 <= 0;
              end
            else
              if (!Ilo)
                begin
                  Iil <= 0;
                  oI0 <= 0;
                end
              else
                if (!oI0 && (I00 == l0o))
                  begin
                    Iil <= 0;
                    oI0 <= 1;
                  end
                else
                  if (!oI0 && (i10 == l0o))
                    begin
                      Iil <= 1;
                      oI0 <= 1;
                    end
                  else
                    if (!oI0 && (li0 == l0o))
                      begin
                        Iil <= 2;
                        oI0 <= 1;
                      end
                    else
                      if (!oI0 && (OI1 == l0o))
                        begin
                          Iil <= 3;
                          oI0 <= 1;
                        end
                      else
                        if (!oI0 && (ol1 == l0o))
                          begin
                            Iil <= 4;
                            oI0 <= 1;
                          end
                        else
                          if (!oI0 && (I11 == l0o))
                            begin
                              Iil <= 5;
                              oI0 <= 1;
                            end
                          else
                            if (!oI0 && (io1 == l0o))
                              begin
                                Iil <= 6;
                                oI0 <= 1;
                              end
                            else
                              if (!oI0 && (lOo == l0o))
                                begin
                                  Iil <= 7;
                                  oI0 <= 1;
                                end
                              else
                                begin
                                  Iil <= Iil;
                                  oI0 <= oI0;
                                end
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              OO0 <= 0;
            else
              if (oI0)
                case (Iil)
                  0:
                    OO0 <= l00;
                  1:
                    OO0 <= Oo0;
                  2:
                    OO0 <= oi0;
                  3:
                    OO0 <= II1;
                  4:
                    OO0 <= il1;
                  5:
                    OO0 <= l11;
                  6:
                    OO0 <= Oi1;
                  7:
                    OO0 <= oOo;
                  default:
                    OO0 <= 0;
                endcase
              else
                OO0 <= 0;
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              ll0 <= 0;
            else
              ll0 <= oI0;
          end
        assign iO0 = { OO0[0], OO0[1], OO0[2], OO0[3], OO0[4], OO0[5], OO0[6], OO0[7] };
      end
  endgenerate
  generate
    if (g_LANE_WIDTH > 1)
      begin
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              begin
                i1l <= 0;
                ool <= 0;
              end
            else
              begin
                i1l <= { L1_DATA_I[0], L1_DATA_I[1], L1_DATA_I[2], L1_DATA_I[3], L1_DATA_I[4], L1_DATA_I[5], L1_DATA_I[6], L1_DATA_I[7] };
                ool <= i1l[6 : 0];
              end
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              begin
                o00 <= 0;
                Io0 <= 0;
                ii0 <= 0;
                lI1 <= 0;
                O01 <= 0;
                o11 <= 0;
                Ii1 <= 0;
                iOo <= 0;
                i00 <= 0;
                lo0 <= 0;
                OO1 <= 0;
                oI1 <= 0;
                I01 <= 0;
                i11 <= 0;
                li1 <= 0;
                OIo <= 0;
              end
            else
              begin
                o00 <= i1l;
                Io0 <= { ool[0], i1l[7 : 1] };
                ii0 <= { ool[1 : 0], i1l[7 : 2] };
                lI1 <= { ool[2 : 0], i1l[7 : 3] };
                O01 <= { ool[3 : 0], i1l[7 : 4] };
                o11 <= { ool[4 : 0], i1l[7 : 5] };
                Ii1 <= { ool[5 : 0], i1l[7 : 6] };
                iOo <= { ool[6 : 0], i1l[7] };
                i00 <= o00;
                lo0 <= Io0;
                OO1 <= ii0;
                oI1 <= lI1;
                I01 <= O01;
                i11 <= o11;
                li1 <= Ii1;
                OIo <= iOo;
              end
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              begin
                lil <= 0;
                iI0 <= 0;
              end
            else
              if (!Ilo)
                begin
                  lil <= 0;
                  iI0 <= 0;
                end
              else
                if (!iI0 && (o00 == l0o))
                  begin
                    lil <= 0;
                    iI0 <= 1;
                  end
                else
                  if (!iI0 && (Io0 == l0o))
                    begin
                      lil <= 1;
                      iI0 <= 1;
                    end
                  else
                    if (!iI0 && (ii0 == l0o))
                      begin
                        lil <= 2;
                        iI0 <= 1;
                      end
                    else
                      if (!iI0 && (lI1 == l0o))
                        begin
                          lil <= 3;
                          iI0 <= 1;
                        end
                      else
                        if (!iI0 && (O01 == l0o))
                          begin
                            lil <= 4;
                            iI0 <= 1;
                          end
                        else
                          if (!iI0 && (o11 == l0o))
                            begin
                              lil <= 5;
                              iI0 <= 1;
                            end
                          else
                            if (!iI0 && (Ii1 == l0o))
                              begin
                                lil <= 6;
                                iI0 <= 1;
                              end
                            else
                              if (!iI0 && (iOo == l0o))
                                begin
                                  lil <= 7;
                                  iI0 <= 1;
                                end
                              else
                                begin
                                  lil <= lil;
                                  iI0 <= iI0;
                                end
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              IO0 <= 0;
            else
              if (iI0)
                case (lil)
                  0:
                    IO0 <= i00;
                  1:
                    IO0 <= lo0;
                  2:
                    IO0 <= OO1;
                  3:
                    IO0 <= oI1;
                  4:
                    IO0 <= I01;
                  5:
                    IO0 <= i11;
                  6:
                    IO0 <= li1;
                  7:
                    IO0 <= OIo;
                  default:
                    IO0 <= 0;
                endcase
              else
                IO0 <= 0;
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              ol0 <= 0;
            else
              ol0 <= iI0;
          end
        assign OI0 = { IO0[0], IO0[1], IO0[2], IO0[3], IO0[4], IO0[5], IO0[6], IO0[7] };
      end
  endgenerate
  generate
    if (g_LANE_WIDTH > 2)
      begin
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              begin
                Ool <= 0;
                iol <= 0;
              end
            else
              begin
                Ool <= { L2_DATA_I[0], L2_DATA_I[1], L2_DATA_I[2], L2_DATA_I[3], L2_DATA_I[4], L2_DATA_I[5], L2_DATA_I[6], L2_DATA_I[7] };
                iol <= Ool[6 : 0];
              end
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              begin
                O10 <= 0;
                oo0 <= 0;
                IO1 <= 0;
                iI1 <= 0;
                l01 <= 0;
                Oo1 <= 0;
                oi1 <= 0;
                IIo <= 0;
                I10 <= 0;
                io0 <= 0;
                lO1 <= 0;
                Ol1 <= 0;
                o01 <= 0;
                Io1 <= 0;
                ii1 <= 0;
                lIo <= 0;
              end
            else
              begin
                O10 <= Ool;
                oo0 <= { iol[0], Ool[7 : 1] };
                IO1 <= { iol[1 : 0], Ool[7 : 2] };
                iI1 <= { iol[2 : 0], Ool[7 : 3] };
                l01 <= { iol[3 : 0], Ool[7 : 4] };
                Oo1 <= { iol[4 : 0], Ool[7 : 5] };
                oi1 <= { iol[5 : 0], Ool[7 : 6] };
                IIo <= { iol[6 : 0], Ool[7] };
                I10 <= O10;
                io0 <= oo0;
                lO1 <= IO1;
                Ol1 <= iI1;
                o01 <= l01;
                Io1 <= Oo1;
                ii1 <= oi1;
                lIo <= IIo;
              end
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              begin
                oil <= 0;
                Ol0 <= 0;
              end
            else
              if (!Ilo)
                begin
                  oil <= 0;
                  Ol0 <= 0;
                end
              else
                if (!Ol0 && (O10 == l0o))
                  begin
                    oil <= 0;
                    Ol0 <= 1;
                  end
                else
                  if (!Ol0 && (oo0 == l0o))
                    begin
                      oil <= 1;
                      Ol0 <= 1;
                    end
                  else
                    if (!Ol0 && (IO1 == l0o))
                      begin
                        oil <= 2;
                        Ol0 <= 1;
                      end
                    else
                      if (!Ol0 && (iI1 == l0o))
                        begin
                          oil <= 3;
                          Ol0 <= 1;
                        end
                      else
                        if (!Ol0 && (l01 == l0o))
                          begin
                            oil <= 4;
                            Ol0 <= 1;
                          end
                        else
                          if (!Ol0 && (Oo1 == l0o))
                            begin
                              oil <= 5;
                              Ol0 <= 1;
                            end
                          else
                            if (!Ol0 && (oi1 == l0o))
                              begin
                                oil <= 6;
                                Ol0 <= 1;
                              end
                            else
                              if (!Ol0 && (IIo == l0o))
                                begin
                                  oil <= 7;
                                  Ol0 <= 1;
                                end
                              else
                                begin
                                  oil <= oil;
                                  Ol0 <= Ol0;
                                end
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              lO0 <= 0;
            else
              if (Ol0)
                case (oil)
                  0:
                    lO0 <= I10;
                  1:
                    lO0 <= io0;
                  2:
                    lO0 <= lO1;
                  3:
                    lO0 <= Ol1;
                  4:
                    lO0 <= o01;
                  5:
                    lO0 <= Io1;
                  6:
                    lO0 <= ii1;
                  7:
                    lO0 <= lIo;
                  default:
                    lO0 <= 0;
                endcase
              else
                lO0 <= 0;
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              il0 <= 0;
            else
              il0 <= Ol0;
          end
        assign II0 = { lO0[0], lO0[1], lO0[2], lO0[3], lO0[4], lO0[5], lO0[6], lO0[7] };
      end
  endgenerate
  generate
    if (g_LANE_WIDTH > 3)
      begin
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              begin
                Iol <= 0;
                Oil <= 0;
              end
            else
              begin
                Iol <= { L3_DATA_I[0], L3_DATA_I[1], L3_DATA_I[2], L3_DATA_I[3], L3_DATA_I[4], L3_DATA_I[5], L3_DATA_I[6], L3_DATA_I[7] };
                Oil <= Iol[6 : 0];
              end
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              begin
                l10 <= 0;
                Oi0 <= 0;
                oO1 <= 0;
                Il1 <= 0;
                i01 <= 0;
                lo1 <= 0;
                OOo <= 0;
                oIo <= 0;
                o10 <= 0;
                Ii0 <= 0;
                iO1 <= 0;
                ll1 <= 0;
                O11 <= 0;
                oo1 <= 0;
                IOo <= 0;
                iIo <= 0;
              end
            else
              begin
                l10 <= Iol;
                Oi0 <= { Oil[0], Iol[7 : 1] };
                oO1 <= { Oil[1 : 0], Iol[7 : 2] };
                Il1 <= { Oil[2 : 0], Iol[7 : 3] };
                i01 <= { Oil[3 : 0], Iol[7 : 4] };
                lo1 <= { Oil[4 : 0], Iol[7 : 5] };
                OOo <= { Oil[5 : 0], Iol[7 : 6] };
                oIo <= { Oil[6 : 0], Iol[7] };
                o10 <= l10;
                Ii0 <= Oi0;
                iO1 <= oO1;
                ll1 <= Il1;
                O11 <= i01;
                oo1 <= lo1;
                IOo <= OOo;
                iIo <= oIo;
              end
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              begin
                iil <= 0;
                Il0 <= 0;
              end
            else
              if (!Ilo)
                begin
                  iil <= 0;
                  Il0 <= 0;
                end
              else
                if (!Il0 && (l10 == l0o))
                  begin
                    iil <= 0;
                    Il0 <= 1;
                  end
                else
                  if (!Il0 && (Oi0 == l0o))
                    begin
                      iil <= 1;
                      Il0 <= 1;
                    end
                  else
                    if (!Il0 && (oO1 == l0o))
                      begin
                        iil <= 2;
                        Il0 <= 1;
                      end
                    else
                      if (!Il0 && (Il1 == l0o))
                        begin
                          iil <= 3;
                          Il0 <= 1;
                        end
                      else
                        if (!Il0 && (i01 == l0o))
                          begin
                            iil <= 4;
                            Il0 <= 1;
                          end
                        else
                          if (!Il0 && (lo1 == l0o))
                            begin
                              iil <= 5;
                              Il0 <= 1;
                            end
                          else
                            if (!Il0 && (OOo == l0o))
                              begin
                                iil <= 6;
                                Il0 <= 1;
                              end
                            else
                              if (!Il0 && (oIo == l0o))
                                begin
                                  iil <= 7;
                                  Il0 <= 1;
                                end
                              else
                                begin
                                  iil <= iil;
                                  Il0 <= Il0;
                                end
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              oO0 <= 0;
            else
              if (Il0)
                case (iil)
                  0:
                    oO0 <= o10;
                  1:
                    oO0 <= Ii0;
                  2:
                    oO0 <= iO1;
                  3:
                    oO0 <= ll1;
                  4:
                    oO0 <= O11;
                  5:
                    oO0 <= oo1;
                  6:
                    oO0 <= IOo;
                  7:
                    oO0 <= iIo;
                  default:
                    oO0 <= 0;
                endcase
              else
                oO0 <= 0;
          end
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              O00 <= 0;
            else
              O00 <= Il0;
          end
        assign lI0 = { oO0[0], oO0[1], oO0[2], oO0[3], oO0[4], oO0[5], oO0[6], oO0[7] };
      end
  endgenerate
  generate
    if (g_LANE_WIDTH == 1)
      begin
        always
          @(posedge CLOCK_I or
            negedge l1l)
          begin
            if (!l1l)
              O0o <= 0;
            else
              O0o <= (ll0) ? 1 : 0;
          end
      end
    else
      if (g_LANE_WIDTH == 2)
        begin
          always
            @(posedge CLOCK_I or
              negedge l1l)
            begin
              if (!l1l)
                O0o <= 0;
              else
                O0o <= (ll0 && ol0) ? 1 : 0;
            end
        end
      else
        if (g_LANE_WIDTH == 4)
          begin
            always
              @(posedge CLOCK_I or
                negedge l1l)
              begin
                if (!l1l)
                  O0o <= 0;
                else
                  O0o <= (ll0 && ol0 && il0 && O00) ? 1 : 0;
              end
          end
  endgenerate
  always
    @(posedge CLOCK_I or
      negedge l1l)
    begin
      if (!l1l)
        begin
          I0o <= 0;
        end
      else
        begin
          I0o[10 : 0] <= { I0o[9 : 0], O0o };
        end
    end
  generate
    if (g_LANE_WIDTH >= 1)
      begin
        cdcfifo #(.ADDR_WIDTH(12), .DATA_WIDTH(8)) o0o(.k_lim(12'h064),
                                                       .wrrstn(l1l),
                                                       .wrclk(CLOCK_I),
                                                       .wrreq(oI0),
                                                       .wrdata(iO0),
                                                       .wrfull(),
                                                       .wrusedw(),
                                                       .rdrstn(l1l),
                                                       .rdclk(CLOCK_I),
                                                       .rdreq(I0o[3]),
                                                       .rddata(L0_DATA_O),
                                                       .rdempty(),
                                                       .rdusedw());
      end
    if (g_LANE_WIDTH >= 2)
      begin
        cdcfifo #(.ADDR_WIDTH(12), .DATA_WIDTH(8)) i0o(.k_lim(12'h064),
                                                       .wrrstn(l1l),
                                                       .wrclk(CLOCK_I),
                                                       .wrreq(iI0),
                                                       .wrdata(OI0),
                                                       .wrfull(),
                                                       .wrusedw(),
                                                       .rdrstn(l1l),
                                                       .rdclk(CLOCK_I),
                                                       .rdreq(I0o[3]),
                                                       .rddata(L1_DATA_O),
                                                       .rdempty(),
                                                       .rdusedw());
      end
    if (g_LANE_WIDTH >= 3)
      begin
        cdcfifo #(.ADDR_WIDTH(12), .DATA_WIDTH(8)) O1o(.k_lim(12'h064),
                                                       .wrrstn(l1l),
                                                       .wrclk(CLOCK_I),
                                                       .wrreq(Ol0),
                                                       .wrdata(II0),
                                                       .wrfull(),
                                                       .wrusedw(),
                                                       .rdrstn(l1l),
                                                       .rdclk(CLOCK_I),
                                                       .rdreq(I0o[3]),
                                                       .rddata(L2_DATA_O),
                                                       .rdempty(),
                                                       .rdusedw());
      end
    if (g_LANE_WIDTH == 4)
      begin
        cdcfifo #(.ADDR_WIDTH(12), .DATA_WIDTH(8)) I1o(.k_lim(12'h064),
                                                       .wrrstn(l1l),
                                                       .wrclk(CLOCK_I),
                                                       .wrreq(Il0),
                                                       .wrdata(lI0),
                                                       .wrfull(),
                                                       .wrusedw(),
                                                       .rdrstn(l1l),
                                                       .rdclk(CLOCK_I),
                                                       .rdreq(I0o[3]),
                                                       .rddata(L3_DATA_O),
                                                       .rdempty(),
                                                       .rdusedw());
      end
  endgenerate
endmodule
