/*****************************************************************************************************************************
--
--    File Name    : bus_cdc_synchornizer.v

--    Description  : This module implements CDC synchronizer for buses
--                   

-- Targeted device : Microsemi-SoC
-- Author          : India Solutions Team

-- SVN Revision Information:
-- SVN $Revision: TBD
-- SVN $Date: TBD
--
--
--
-- COPYRIGHT 2016 BY MICROSEMI 
-- THE INFORMATION CONTAINED IN THIS DOCUMENT IS SUBJECT TO LICENSING RESTRICTIONS 
-- FROM MICROSEMI CORP.  IF YOU ARE NOT IN POSSESSION OF WRITTEN AUTHORIZATION FROM 
-- MICROSEMI FOR USE OF THIS FILE, THEN THE FILE SHOULD BE IMMEDIATELY DESTROYED AND 
-- NO BACK-UP OF THE FILE SHOULD BE MADE. 
-- 

--******************************************************************************************************************************/

`timescale 1 ns / 1 ps
module bus_cdc_synchornizer_hdl 
#(
      parameter g_BUS_WIDTH = 32
)(

      input wire DEST_CLOCK_I,      

      input wire [(g_BUS_WIDTH - 1):0] input_bus_i,
      output reg [(g_BUS_WIDTH - 1):0] output_bus_o
);
  
  wire [(g_BUS_WIDTH - 1):0] input_bus_gray;
  reg [(g_BUS_WIDTH - 1):0] gray_bus_reg1, gray_bus_reg2;
  integer i;

  assign input_bus_gray = (input_bus_i >> 1) ^ input_bus_i; /* Binary to Gray conversion */

  always
    @(posedge DEST_CLOCK_I)
    begin    
      gray_bus_reg1  <= input_bus_gray;
      gray_bus_reg2  <= gray_bus_reg1;
    end
    
    /* Gray to Binary conversion */
  always
    @(gray_bus_reg2)
    begin
      for (i = 0; i < g_BUS_WIDTH; i = i + 1)
        begin
          output_bus_o[i] = ^(gray_bus_reg2 >> i);
        end
    end



endmodule